/************************************************************************
 FILE: CubeIt0.cpp

 simplest cubeit solution program

 makes a world, allows cubes to be created and moved around
 gives camera controls

 does everything with sliders - easy to write, maybe not as easy to
 use

 written 10/21/2007 by Mike Gleicher
 
     CVS Info: Date checked in: $Date: 2007/10/22 15:38:30 $
               Revision number: $Revision: 1.4 $

***********************************************************************/

#include "stdio.h"
#include "math.h"
#include <windows.h>

// the FlTk headers have lots of warnings - these are bad, but there's not
// much we can do about them
#pragma warning(push)
#pragma warning(disable:4311)		// convert void* to long
#pragma warning(disable:4312)		// convert long to void*
#include <FL/Fl_Gl_Window.h>
#include <Fl/Fl.h>
#include <Fl/Fl_Multiline_Output.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <Fl/Fl_Double_Window.h>
#include <Fl/Fl_Button.h>
#include <FL/FL_Roller.H>
#include <FL/Fl_Value_Slider.H>
#pragma warning(pop)

#include "3DUtils.H"
#include "CubeIt.H"
#include "RunButton.H"

// cheesy camera controls - have sliders for each camera param
// make them global variables for easy access - when we draw, the
// camera just looks at them
Fl_Value_Slider* sl_eyeX = 0;
Fl_Value_Slider* sl_eyeY = 0;
Fl_Value_Slider* sl_eyeZ = 0;
Fl_Value_Slider* sl_rotX = 0;
Fl_Value_Slider* sl_rotY = 0;
Fl_Value_Slider* sl_rotZ = 0;
Fl_Value_Slider* sl_fov = 0;

// all the camera widgets do is damage the main window when they
// change (we'll use the extra parameter to store the main window)
void camSliderCB(Fl_Widget*, void * w)
{
  ((Fl_Widget*) w)->damage(1);
}

// make an Fl_Value_Slider and set all of the parameters correctly
// makes a lot of assumptions of where we're going to put things
Fl_Value_Slider* makeValueSlider(const int xPos, const char* label, 
								 const double minv, const double maxv,
								 Fl_Widget* toDamage,
								 const double initv=0)
{
  Fl_Value_Slider* sl = new Fl_Value_Slider(xPos, 10, 30, 150, label);
  sl->minimum(minv);
  sl->maximum(maxv);
  if (maxv-minv > 45)
	sl->step(1);
  else
	sl->step(.1);
  sl->value(initv);
  sl->textsize(12);
  sl->callback(camSliderCB,toDamage);
  return sl;
}

//*********************************************************************************
//
// Cheesy UI!
// Need a callback for each dimension
// and a function to make the rollers
// Forward declare here, and put the code at the end
void makeRollers(int x, int y, int h, Fl_Widget* w);
void xRollerCB(Fl_Widget* w, void* p);
void yRollerCB(Fl_Widget* w, void* p);
void zRollerCB(Fl_Widget* w, void* p);

// cheesy UI - have buttons to cycle through the selection
// need a callback that causes the selection to change
Fl_Button* but_next = 0;
Fl_Button* but_prev = 0;
// callback - make sure that the parameter gets the window to damage!
void nextButtonCallback(Fl_Widget*, void* w)
{
	selectedCube ++;
	selectedCube = selectedCube % (int) theCubes.size();
	((Fl_Widget*) w)->damage(1);
}
void prevButtonCallback(Fl_Widget*, void* w)
{
	selectedCube --;
	if (selectedCube < 0) selectedCube = static_cast<int>(theCubes.size())-1;
	((Fl_Widget*) w)->damage(1);
}


//*************************************************************************
//
// Create a widget that is a GL Window that has the world in it
//
// If you remember from the example program, all of the work happens
// (or can happen) in a Widget class that has draw and handle
// functions. In this example, I break those two "standard" routines
// into a bunch of pieces to make the code more organized
//
//*************************************************************************
class MyGlWindow : public Fl_Gl_Window {
	public:
		MyGlWindow(int x, int y, int w, int h);
		// we need to know what time it is for making the moving object
		Fl_Slider* time;
	private:
		void draw();					// standard FlTk

		// Put things to set up the projection into a method so we can use
		// it places other than draw
		void setProjection();
};

//**************************************************************************
//
// Implementation of MyGLWindow
//==========================================================================
MyGlWindow::MyGlWindow(int x, int y, int w, int h) 
:  Fl_Gl_Window(x,y,w,h), time(0)
//==========================================================================
{
	// note: we will use some fancy GL features, so we have to set the
	// mode correctly
	// specifically, we need the stencil buffer (so we can draw shadows
	// only on the ground plane) and alpha blending (to have transparent
	// shadows)
 mode( FL_RGB|FL_ALPHA|FL_DOUBLE | FL_STENCIL );
}

//**************************************************************************
//
// draw the windows proceeds as follows:
// first, clear the window (be sure to clear the Z-buffer)
// then set up the projection and view
// then draw the ground plane
// then set up for shadows and draw the objects
//		which causes their shadows to be drawn
// then draw the objects
//==========================================================================
void MyGlWindow::draw(void)
//==========================================================================
{

  // clear the window, be sure to clear the Z-Buffer too
  glClearColor(0,0,.3f,0);		// background should be blue

  // we need to clear out the stencil buffer since we'll use
  // it for shadows
  glClearStencil(0);

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    
  // set up the drawing transforms
  setProjection();

  // now draw the ground plane
  setupFloor();
  drawFloor();

  // now to draw the shadows
  setupShadows();

  // Draw the objects without setting the colors (otherwise we'd
  // get colored shadows
  if (time) {
	  glPushMatrix();
	  glRotated(time->value()*4.0,0,1,0);
	  drawCube(5,1,0,2);
	  glPopMatrix();
  }
  drawCubes(0);
  unsetupShadows();

  // Finally, draw the objects
  drawCubes();

  // make the thing that goes around
  if (time) {
	  glColor3f(0,1,0);
	  glPushMatrix();
	  glRotated(time->value()*4.0,0,1,0);
	  drawCube(5,1,0,2);
	  drawCube(5,1,-1.5,1);
	  glPopMatrix();
  }
}

//**************************************************************************
//
// * Set up the view projection matrix which is useful for selection
//==========================================================================
void MyGlWindow::setProjection(void)
//==========================================================================
{
  glMatrixMode(GL_PROJECTION);
  glViewport(0,0,w(),h());
  glLoadIdentity();

  // Compute the aspect ratio so we don't distort things
  double aspect = ((double) w()) / ((double) h());
  gluPerspective(sl_fov->value(), aspect, 1, 100);

  // Put the camera where we want it to be
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  // translate and rotate the camera
  glTranslated(sl_eyeX->value(), sl_eyeY->value(), sl_eyeZ->value());
  glRotated(sl_rotX->value(),1,0,0);
  glRotated(sl_rotY->value(),0,1,0);
  glRotated(sl_rotZ->value(),0,0,1);
}


// the main routine makes the window, and then runs an even loop
// until the window is closed
// the most complicated thing it does is create all of the widgets
// for the UI
int main(int, char**)
{
	Fl_Double_Window* wind = new Fl_Double_Window(100,100,840,500,"GL 3D Sample");

	wind->begin();		// put widgets inside of the window
	  // the "main" window
	  MyGlWindow* gl = new MyGlWindow(10,10,480,480);
	  // the sliders for the camera
	  sl_eyeX = makeValueSlider(500,"eyeX",-20,20,gl);
	  sl_eyeY = makeValueSlider(540,"eyeY",-20,20,gl);
	  sl_eyeZ = makeValueSlider(580,"eyeZ",-20,20,gl,-20);
	  sl_rotX = makeValueSlider(660,"rotX",-180,180,gl,25);
	  sl_rotY = makeValueSlider(700,"rotY",-180,180,gl);
	  sl_rotZ = makeValueSlider(740,"rotZ",-180,180,gl);
	  sl_fov  = makeValueSlider(800,"fov",20,90,gl,60);
	  // put up the rollers
	  makeRollers(580,200,100,gl);
	  // buttons for selection
	  but_next = new Fl_Button(500,240,30,30,"@>");
	  but_next->callback(nextButtonCallback,gl);
	  but_prev = new Fl_Button(540,240,30,30,"@<");
	  but_prev->callback(prevButtonCallback,gl);

	  // a run button
	  RunSlider* sl = new RunSlider(gl,90,560,460,300,20);
	  RunButton* rb = new RunButton(sl,   500,460, 50,20);
	  gl->time = sl;
	wind->end();


	wind->show();	// this actually opens the window

	// make some cubes to show
	theCubes.push_back(Cube(-3,3, 3));
	theCubes.push_back(Cube( 3,3, 3));
	theCubes.push_back(Cube(-3,3,-3));
	theCubes.push_back(Cube( 3,3,-3));
	// put some cubes over space so we can test the shadows only cast
	// onto the floor (not onto space)
	theCubes.push_back(Cube(7,3,0,1));
	theCubes.push_back(Cube(-7,3,0,1));
	theCubes.push_back(Cube(0,3,7,1));
	theCubes.push_back(Cube(0,3,-7,1));
	// put a cube over another cube so we can see that the shadows
	// don't stack
	theCubes.push_back(Cube( 0,4,-4,2));
	theCubes.push_back(Cube( 0,2,-3,1));

	// become interactive
	Fl::run();
	delete wind;

	return 1;
}

// actual implementation of the cheesy UI

//*********************************************************************************
//
// Cheesy UI!
// Need a callback for each dimension
// the reseting each time doesn't really work, but its a cheesy
// way to avoid having to reset the rollers when a new object is selected
//
//*********************************************************************************

//*********************************************************************************
//
// * Callback for X dimension
//=================================================================================
void xRollerCB(Fl_Widget* w, void* p)
//=================================================================================
{
  if (selectedCube > -1) {
	theCubes[selectedCube].x += (float) static_cast<Fl_Roller*>(w)->value();
	static_cast<Fl_Roller*>(w)->value(0);
	if (p) static_cast<Fl_Widget*>(p)->damage(1);
  }
}
//*********************************************************************************
//
// * Callback for Y dimension
//=================================================================================
void yRollerCB(Fl_Widget* w, void* p)
//=================================================================================
{
  if (selectedCube > -1) {
	theCubes[selectedCube].y += (float) static_cast<Fl_Roller*>(w)->value();
	static_cast<Fl_Roller*>(w)->value(0);
	if (p) static_cast<Fl_Widget*>(p)->damage(1);
  }
}
//*********************************************************************************
//
// * Callback for Z dimension
//=================================================================================
void zRollerCB(Fl_Widget* w, void* p)
//=================================================================================
{
  if (selectedCube > -1) {
	theCubes[selectedCube].z += (float) static_cast<Fl_Roller*>(w)->value();
	static_cast<Fl_Roller*>(w)->value(0);
	if (p) static_cast<Fl_Widget*>(p)->damage(1);
  }
}

//*********************************************************************************
// a cheesy UI
//
// Use FlTk Rollers to move the selected cube
//
// I don't recommend this interface - its just here to show you that
// there are simple things that work
//
// Make rollers - be sure you're inside of an FlTk Window
//                give it an xy position and size for the rollers
//                also give it an fltk widget to damage (your window)
//                anytime something changes
//*********************************************************************************
void makeRollers(int x, int y, int h, Fl_Widget* w)
{
  Fl_Roller* rx = new Fl_Roller(x,y,25,h,"X");
  rx->minimum(-1);
  rx->maximum(1);
  rx->callback(xRollerCB,w);
  Fl_Roller* ry = new Fl_Roller(x+40,y,25,h,"Y");
  ry->minimum(-1);
  ry->maximum(1);
  ry->callback(yRollerCB,w);
  Fl_Roller* rz = new Fl_Roller(x+80,y,25,h,"Z");
  rz->minimum(-1);
  rz->maximum(1);
  rz->callback(zRollerCB,w);
}


